using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using Avalonia;
using Avalonia.Controls;
using Avalonia.Input;
using Avalonia.Interactivity;
using Avalonia.Markup.Xaml;
using Avalonia.Threading;
using Microsoft.AspNetCore.SignalR.Client;

namespace GUI
{
    public class MainWindow : Window
    {
        private readonly HubConnection _connection;
        private string _gameId;
        private readonly Dictionary<int, TetrisGrid> _grids = new();
        private bool _multiplayer;
        private int _ownGridSeed;
        private readonly Panel _panel;
        private Timer _timer;

        public MainWindow()
        {
            InitializeComponent();
#if DEBUG
            this.AttachDevTools();
#endif
            _panel = this.Find<Panel>("Panel");

            string url = "https://tetris.rubend.nl/TetrisHub"; //Custom deployment URL.

            // De Builder waarmee de connectie aangemaakt wordt:
            _connection = new HubConnectionBuilder()
                .WithUrl(url)
                .WithAutomaticReconnect()
                .Build();
            _connection.On<int, string>("Seed", (seed, name) =>
            {
                Dispatcher.UIThread.InvokeAsync(()=>
                {
                    _multiplayer = true;
                    _ownGridSeed = seed;
                    this.Find<StackPanel>("JoinPanel").IsVisible = false;
                    this.Find<Button>("StartButton").IsVisible = true;
                    AddGrid(seed, name);
                });
            });
            // De eerste paramater moet gelijk zijn met de methodenaam in TetrisHub.cs
            // Wat er tussen de <..> staat bepaald wat de type van de paramater `seed` is.
            // Op deze manier loopt het onderstaande gelijk met de methode in TetrisHub.cs.
            _connection.On<int, string>("Join",
                (seed, name) => { Dispatcher.UIThread.InvokeAsync(() => AddGrid(seed, name)); });
            _connection.On<int, string>("KeyPress",
                (seed, direction) => { Dispatcher.UIThread.InvokeAsync(() => _grids[seed].KeyPress(direction)); });
            _connection.On("Timer", () => MoveDown());
            _connection.On("AlreadyStarted", () => AlreadyStarted());

            // Let op: het starten van de connectie moet *nadat* alle event listeners zijn gezet!
            // Als de methode waarin dit voorkomt al `async` (asynchroon) is, dan kan `Task.Run` weggehaald worden.
            // In het startersproject staat dit in de constructor, daarom is dit echter wel nodig:
            Task.Run(async () => await _connection.StartAsync());
        }

        private void AlreadyStarted()
        {
            Task.Delay(TimeSpan.FromMilliseconds(1000))
                .ContinueWith(task => Dispatcher.UIThread.InvokeAsync(()=>this.Find<Border>("AlreadyStarted").IsVisible=false));
            Dispatcher.UIThread.InvokeAsync(()=>this.Find<Border>("AlreadyStarted").IsVisible=true);
        }
        private void MoveDown(object? nothing = null)
        {
            Dispatcher.UIThread.InvokeAsync(() => this.Find<Button>("StartButton").IsVisible = false);
            foreach (var entry in _grids) entry.Value.MoveDown();
        }

        private void AddGrid(int seed, string name)
        {
            TetrisGrid tetrisGrid = new();
            tetrisGrid.SetSeed(seed);
            tetrisGrid.SetName(name);
            _panel.Children.Add(tetrisGrid);
            _grids.Add(seed, tetrisGrid);
        }

        private async void JoinClicked(object sender, RoutedEventArgs e)
        {
            _gameId = this.Find<TextBox>("GameId").Text ?? "EMPTY";
            string name = this.Find<TextBox>("Name").Text ?? "No Name";
            await _connection.InvokeAsync("Join", _gameId, name);
        }

        private async void StartClicked(object sender, RoutedEventArgs e)
        {
            if (_multiplayer) await _connection.InvokeAsync("Start", _gameId);
            else _timer = new Timer(MoveDown, null, 0, 500);
            this.Find<Button>("StartButton").IsVisible = false;
        }

        private void SinglePlayerClicked(object sender, RoutedEventArgs e)
        {
            _ownGridSeed = Guid.NewGuid().GetHashCode();
            AddGrid(_ownGridSeed, "No Name");
            _multiplayer = false;
            this.Find<StackPanel>("JoinPanel").IsVisible = false;
            this.Find<Button>("StartButton").IsVisible = true;
        }

        private void InitializeComponent()
        {
            AvaloniaXamlLoader.Load(this);
        }


        private void Panel_OnKeyDown(object? sender, KeyEventArgs e)
        {
            if (_ownGridSeed == 0) return;
            if (_grids[_ownGridSeed].Lost) return;
            if (_multiplayer) _connection.InvokeAsync("KeyPress", _gameId, _ownGridSeed, e.Key.ToString());
            else _grids[_ownGridSeed].KeyPress(e.Key.ToString());
        }
    }
}