using System;
using System.Collections.Generic;
using System.Linq;
using Avalonia.Controls;
using Avalonia.Markup.Xaml;
using Avalonia.Threading;

namespace GUI
{
    public class TetrisGrid : UserControl
    {
        public static int GridHeight = 16;
        public static int GridWidth = 10;
        private readonly Grid _holdGrid;
        private readonly Grid _nextGrid;
        private readonly ScoreBox _scoreBox;
        private readonly Grid _tetrisGrid;
        private Bag _bag;
        private bool _hasHold;
        private Matrix? _holdMatrix;
        private Piece _lastPlaced;
        private Matrix _nextMatrix;
        public int Level = 0; //TODO option to change this.
        public bool Lost;

        public TetrisGrid()
        {
            InitializeComponent();
            _tetrisGrid = this.Find<Grid>("TetrisGrid");
            _tetrisGrid.RowDefinitions =
                new RowDefinitions(string.Join(",", Enumerable.Repeat("25", GridHeight).ToArray()));
            _tetrisGrid.ColumnDefinitions =
                new ColumnDefinitions(string.Join(",", Enumerable.Repeat("25", GridWidth).ToArray()));
            _nextGrid = this.Find<Grid>("NextGrid");
            _nextGrid.RowDefinitions = new RowDefinitions(string.Join(",", Enumerable.Repeat("25", 4).ToArray()));
            _nextGrid.ColumnDefinitions = new ColumnDefinitions(string.Join(",", Enumerable.Repeat("25", 4).ToArray()));
            _holdGrid = this.Find<Grid>("HoldGrid");
            _holdGrid.RowDefinitions = new RowDefinitions(string.Join(",", Enumerable.Repeat("25", 4).ToArray()));
            _holdGrid.ColumnDefinitions = new ColumnDefinitions(string.Join(",", Enumerable.Repeat("25", 4).ToArray()));
            _scoreBox = this.Find<ScoreBox>("ScoreBox");
            _scoreBox.Show();
        }

        public void SetSeed(int seed)
        {
            _bag = new Bag(seed);
            new Piece(_nextGrid, _bag.GetNextMatrix(), 0, 0, false, false).Place();
            _nextMatrix = _bag.GetNextMatrix();
            NewTetris(init:true);
        }

        public void SetName(string name)
        {
            _scoreBox.SetName(name);
        }

        public void MoveDown()
        {
            if (Lost) return;
            Action moveDown = () =>
            {
                if (_lastPlaced.Move(0, 1)) return;
                _lastPlaced.Ghost.Remove();
                _hasHold = false;
                NewTetris();
            };
            Dispatcher.UIThread.InvokeAsync(moveDown);
        }

        private void InitializeComponent()
        {
            AvaloniaXamlLoader.Load(this);
        }

        private void CheckFullLines()
        {
            List<int> rows = _tetrisGrid.Children.AsEnumerable()
                .Select(item => (Control) item)
                .Select(Grid.GetRow)
                .GroupBy(item => item)
                .Where(item => item.Count() == GridWidth)
                .Select(item => item.Key)
                .OrderBy(item => item)
                .ToList();
            if (rows.Count == 0) return;
            //score calculation: https://tetris.wiki/Scoring#Recent_guideline_compatible_games
            _scoreBox.AddRows(rows.Count);
            foreach (
                IControl item in _tetrisGrid.Children.ToList().AsEnumerable() //don't question it.
                    .Where(item => rows.Contains(Grid.GetRow((Control) item)))
            ) _tetrisGrid.Children.Remove(item);

            foreach (
                Control item in _tetrisGrid.Children
                    .Select(item => (Control) item)
                    .Where(item => rows[0] > Grid.GetRow(item))
            ) Grid.SetRow(item, Grid.GetRow(item) + rows.Count);
        }

        private void NewTetris(Matrix? next = null, bool init=false)
        {
            CheckFullLines();
            _lastPlaced = new Piece(_tetrisGrid, next ?? _nextMatrix, GridWidth / 2 - 2, 0);
            if (!_lastPlaced.Place())
            {
                Lost = true;
                Classes.Add("lost");
            }

            if (next.HasValue) return;
            _nextMatrix = _bag.GetNextMatrix();
            _nextGrid.Children.Clear();
            new Piece(_nextGrid, _nextMatrix, 0, 0, false, false).Place();
        }

        private void Hold()
        {
            if (_hasHold) return;
            _hasHold = true;
            _lastPlaced.Remove();
            var oldLastPlaced = _lastPlaced.Matrix;
            _holdGrid.Children.Clear();
            if (_holdMatrix.HasValue) NewTetris(_holdMatrix);
            else NewTetris();
            _holdMatrix = oldLastPlaced;
            Piece holdPiece = new(_holdGrid, _holdMatrix.Value, 0, 0, false, false);
            holdPiece.Place();
        }

        public void KeyPress(string key)
        {
            switch (key)
            {
                case "Left":
                    _lastPlaced.Move(-1, 0);
                    break;
                case "Right":
                    _lastPlaced.Move(1, 0);
                    break;
                case "Z":
                    _lastPlaced.Rotate(-1);
                    break;
                case "X":
                    _lastPlaced.Rotate(1);
                    break;
                case "C":
                    Hold();
                    break;
                case "Down":
                    if (_lastPlaced.Move(0, 1)) _scoreBox.Add(_lastPlaced.Matrix.CountBlocks());
                    break;
            }
        }
    }
}